import numpy as np
from scipy import stats
from ._test import MantelResult

try:
    import matplotlib.pyplot as plt
    from matplotlib.axes import Axes
except ImportError:
    plt = None


def plot(
    result,
    axis=None,
    alpha=0.05,
    hist_color="lightgray",
    gaussian_color="black",
    acceptance_color="black",
    rejection_color="black",
):
    """
    Plot a histogram of the sample correlations from a MantelResult object on
    a Matplotlib axis. If no axis is provided, a new plot is created and
    returned. The theoretical normal distribution (and confidence interval)
    is also shown, along with the veridical correlation.

    Parameters
    ----------
    result : MantelResult
        The reuslt object output from mantel.test()
    axis : matplotlib.axes.Axes
        The matplotlib figure to draw on.
    alpha : float
        Significance level for rejecting the null hypothesis (default: 5%)
    hist_color: str
        Color used for the histogram bars (default: 'lightgray').
    gaussian_color: str
        Color used for the normal distribution curve and the confidence
        interval limits (default: 'black').
    acceptance_color: str
        Color used for drawing the vertical line and the label of the
        veridical correlation if the null hypothesis is rejected according to
        the significance level value (default: 'black').
    rejection_color: str
        Color used for drawing the vertical line and the label of the
        veridical correlation if the null hypothesis cannot be rejected
        according to the significance level value (default: 'black').

    Returns
    -------
    Matplotlib Figure and Axis objects
    """
    if plt is None:
        raise ImportError("Matplotlib is required for plotting")
    if not isinstance(result, MantelResult):
        raise TypeError("result shoud be a MantelResult object")
    if axis is None:
        fig, axis = plt.subplots()
    elif isinstance(axis, Axes):
        fig = axis.get_figure()
    else:
        raise TypeError("axis should be a Matplotlib Axis object")

    lower = max(-5 * result.std + result.mean, -1)
    upper = min(5 * result.std + result.mean, 1)

    n_bins = 100 if result.perms >= 1000 else 20
    axis.hist(
        result.correlations,
        bins=n_bins,
        range=(lower, upper),
        density=True,
        histtype="stepfilled",
        color=hist_color,
    )

    x = np.linspace(lower, upper, 200)
    y = stats.norm.pdf(x, result.mean, result.std)
    axis.plot(x, y, color=gaussian_color)

    min_corr, max_corr = confidence_interval(result, alpha)
    axis.vlines(
        x=[min_corr, max_corr],
        ymin=0,
        ymax=[
            stats.norm.pdf(min_corr, result.mean, result.std),
            stats.norm.pdf(max_corr, result.mean, result.std),
        ],
        linestyle="-",
        color=gaussian_color,
    )

    threshold_color = (
        rejection_color if min_corr <= result.r <= max_corr else acceptance_color
    )
    axis.axvline(x=result.r, linestyle=":", color=threshold_color)
    axis.annotate("{:.2f}".format(result.r), xy=(result.r, 0.9), color=threshold_color)

    axis.set_xlim(lower, upper)
    axis.set_xlabel("Correlation coefficient")
    axis.set_yticks([])
    return fig, axis


def confidence_interval(result, alpha=0.05):
    """
    Return the confidence interval in a normal distribution(characterized by
    its given mean and its given standard deviation) for rejecting the null
    hypothesis with some given significance level and some given tail
    method.

    Parameters
    ----------
    result : float
        MantelResult object
    alpha : float (between 0 and 1), optional
        The probability of rejecting the null hypothesis when it is true
        (default: 0.05).

    Returns
    -------
    lower_bound : float
        Lower bound of the confidence interval.
    upper_bound : float
        Upper bound of the confidence interval.
    """
    if not (0 < alpha < 1):
        raise ValueError("The alpha level must be in the interval (0, 1)")
    if result.tail == "upper":
        lower_bound = stats.norm.ppf(1e-10, result.mean, result.std)
        upper_bound = stats.norm.ppf(1 - alpha, result.mean, result.std)
    elif result.tail == "lower":
        lower_bound = stats.norm.ppf(alpha, result.mean, result.std)
        upper_bound = stats.norm.ppf(1 - 1e-10, result.mean, result.std)
    elif result.tail == "two-tail":
        lower_bound = stats.norm.ppf(alpha / 2, result.mean, result.std)
        upper_bound = stats.norm.ppf(1 - alpha / 2, result.mean, result.std)
    return lower_bound, upper_bound



import math
from itertools import permutations
import numpy as np
from scipy import spatial, stats


class MantelResult:
    """
    Object representing the result of a Mantel test. Specific properties can
    be queried by dot notation (e.g. `result.correlations`, `result.r`,
    `result.p`, etc.) or the result object can be treated like a tuple of the
    form `(r, p, z)` for backwards compatibility.
    """

    def __init__(self, correlations, method, tail, ignore_nans):
        self._correlations = correlations
        self._perms = len(correlations)
        self._method = method
        self._tail = tail
        self._ignore_nans = ignore_nans
        self._mean = None
        self._std = None
        self._p = None

    def __repr__(self):
        return f"MantelResult({self.r}, {self.p}, {self.z})"

    def __iter__(self):
        yield self.r
        yield self.p
        yield self.z

    def __getitem__(self, index):
        if index == 0:
            return self.r
        if index == 1:
            return self.p
        if index == 2:
            return self.z
        raise IndexError("index out of range")

    @property
    def perms(self):
        return self._perms

    @property
    def method(self):
        return self._method

    @property
    def tail(self):
        return self._tail

    @property
    def ignore_nans(self):
        return self._ignore_nans

    @property
    def correlations(self):
        return self._correlations

    @property
    def mean(self):
        if self._mean is None:
            self._mean = np.mean(self.correlations)
        return self._mean

    @property
    def std(self):
        if self._std is None:
            self._std = np.std(self.correlations)
        return self._std

    @property
    def r(self):
        return self.correlations[0]

    @property
    def p(self):
        if self._p is None:
            if self.tail == "upper":
                self._p = sum(self.correlations >= self.r) / self.perms
            elif self.tail == "lower":
                self._p = sum(self.correlations <= self.r) / self.perms
            else:
                self._p = sum(abs(self.correlations) >= abs(self.r)) / self.perms
        return self._p

    @property
    def z(self):
        return (self.r - self.mean) / self.std


def test(X, Y, perms=10000, method="pearson", tail="two-tail", ignore_nans=False):
    """
    Takes two distance matrices (either redundant matrices or condensed vectors)
    and performs a Mantel test. The Mantel test is a significance test of the
    correlation between two distance matrices.

    Parameters
    ----------
    X : array_like
        First distance matrix (condensed or redundant).
    Y : array_like
        Second distance matrix (condensed or redundant), where the order of
        elements corresponds to the order of elements in the first matrix.
    perms : int, optional
        The number of permutations to perform (default: 10000). A larger
        number gives more reliable results but takes longer to run. If the
        number of possible permutations is smaller, all permutations will
        be tested. This can be forced by setting perms to 0.
    method : str, optional
        Type of correlation coefficient to use; either 'pearson' or 'spearman'
        (default: 'pearson').
    tail : str, optional
        Which tail to test in the calculation of the empirical p-value; either
        'upper', 'lower', or 'two-tail' (default: 'two-tail').
    ignore_nans : bool, optional
        Ignore NaN values in the Y matrix (default: False). This can be
        useful if you have missing values in one of the matrices.

    Returns
    -------
    MantelResult object from which various properties can be queried:

    - MantelResult.r - veridical correlation coefficient
    - MantelResult.p - empirical p-value
    - MantelResult.z - standard score
    - MantelResult.correlations - the sample correlations
    - MantelResult.mean - mean of the sample correlations
    - MantelResult.std - standard deviation of the sample correlations
    """

    # Ensure that X and Y are represented as Numpy arrays.
    X = np.asarray(X)
    Y = np.asarray(Y)

    # Check that X and Y are valid distance matrices.
    if (
        spatial.distance.is_valid_dm(np.nan_to_num(X)) == False
        and spatial.distance.is_valid_y(X) == False
    ):
        raise ValueError("X is not a valid condensed or redundant distance matrix")
    if (
        spatial.distance.is_valid_dm(np.nan_to_num(Y)) == False
        and spatial.distance.is_valid_y(Y) == False
    ):
        raise ValueError("Y is not a valid condensed or redundant distance matrix")

    # If X or Y is a redundant distance matrix, reduce it to a condensed distance matrix.
    if len(X.shape) == 2:
        X = spatial.distance.squareform(X, force="tovector", checks=False)
    if len(Y.shape) == 2:
        Y = spatial.distance.squareform(Y, force="tovector", checks=False)

    # Check for size equality.
    if len(X) != len(Y):
        raise ValueError("X and Y are not of equal size")

    # Check for minimum size.
    if len(X) < 3:
        raise ValueError("X and Y should represent at least 3 objects")

    # Check finiteness of X and Y
    if not np.isfinite(X).all():
        raise ValueError(
            "X cannot contain NaNs (but Y may contain NaNs, so consider reordering X and Y)"
        )
    finite_Y = np.isfinite(Y)
    if not ignore_nans and not finite_Y.all():
        raise ValueError('Y may contain NaNs, but "ignore_nans" must be set to True')
    if ignore_nans and finite_Y.all():
        ignore_nans = False  # ignore_nans is True but Y contains no nans

    # If Spearman correlation is requested, convert X and Y to ranks.
    method = method.lower()
    if method == "spearman":
        X, Y = stats.rankdata(X), stats.rankdata(Y)
        Y[~finite_Y] = np.nan  # retain any nans, so that these can be ignored later

    # Check for valid method parameter.
    elif method != "pearson":
        raise ValueError('The method should be set to "pearson" or "spearman"')

    # Check for valid tail parameter.
    tail = tail.lower()
    if tail not in ["upper", "lower", "two-tail"]:
        raise ValueError('The tail should be set to "upper", "lower", or "two-tail"')

    # Now we're ready to start the Mantel test using a number of optimizations:
    #
    # 1. Rather than compute correlation coefficients, we'll just compute the
    #    covariances. This works because the denominator in the equation for the
    #    correlation coefficient will yield the same result however the objects
    #    are permuted, making it redundant. Removing the denominator leaves us
    #    with the covariance.
    #
    # 2. Rather than permute the Y distances and derive the residuals to calculate
    #    the covariance with the X distances, we'll represent the Y residuals in
    #    the matrix and shuffle those directly.
    #
    # 3. If the number of possible permutations is less than the number of
    #    permutations that were requested, we'll run a deterministic test where
    #    we try all possible permutations rather than sample the permutation
    #    space. This gives a faster, deterministic result.

    # Calculate the X and Y residuals, which will be used to compute the
    # covariance under each permutation.
    X_residuals = X - np.mean(X[finite_Y])
    Y_residuals = Y - np.mean(Y[finite_Y])

    # Expand the Y residuals to a redundant matrix.
    Y_residuals_as_matrix = spatial.distance.squareform(
        Y_residuals, force="tomatrix", checks=False
    )

    m = len(Y_residuals_as_matrix)  # number of objects
    n = math.factorial(m)  # number of possible matrix permutations

    # If the number of requested permutations is greater than the number of
    # possible permutations (m!) or the perms parameter is set to 0, then run a
    # deterministic Mantel test
    if perms >= n or perms == 0:
        if ignore_nans:
            correlations = deterministic_test_with_nans(m, n, X, Y_residuals_as_matrix)
        else:
            correlations = deterministic_test(m, n, X_residuals, Y_residuals_as_matrix)
        # correlations[0] is the veridical correlation

    else:
        if ignore_nans:
            correlations = stochastic_test_with_nans(m, perms, X, Y_residuals_as_matrix)
        else:
            correlations = stochastic_test(m, perms, X_residuals, Y_residuals_as_matrix)
        correlations[0] = sum(X_residuals[finite_Y] * Y_residuals[finite_Y]) / np.sqrt(
            sum(X_residuals[finite_Y] ** 2) * sum(Y_residuals[finite_Y] ** 2)
        )  # compute veridical correlation and place in positon 0

    return MantelResult(correlations, method, tail, ignore_nans)


def deterministic_test(m, n, X_residuals, Y_residuals_as_matrix):
    Y_residuals_permuted = np.zeros((m**2 - m) // 2)
    covariances = np.zeros(n)
    for i, order in enumerate(permutations(range(m))):
        Y_residuals_as_matrix_permuted = Y_residuals_as_matrix[order, :][:, order]
        spatial.distance._distance_wrap.to_vector_from_squareform_wrap(
            Y_residuals_as_matrix_permuted, Y_residuals_permuted
        )
        covariances[i] = (X_residuals * Y_residuals_permuted).sum()
    denominator = np.sqrt(sum(X_residuals**2) * sum(Y_residuals_permuted**2))
    return covariances / denominator


def deterministic_test_with_nans(m, n, X, Y_residuals_as_matrix):
    Y_residuals_permuted = np.zeros((m**2 - m) // 2)
    correlations = np.zeros(n)
    for i, order in enumerate(permutations(range(m))):
        Y_residuals_as_matrix_permuted = Y_residuals_as_matrix[order, :][:, order]
        spatial.distance._distance_wrap.to_vector_from_squareform_wrap(
            Y_residuals_as_matrix_permuted, Y_residuals_permuted
        )
        # Since on each permutation we will be ignoring different values in X,
        # the X_residuals need to be recomputed each time depending on which
        # values in permuted Y are finite.
        finite_Y_permuted = np.isfinite(Y_residuals_permuted)
        reduced_X = X[finite_Y_permuted]
        reduced_X_residuals = reduced_X - reduced_X.mean()
        reduced_Y_residuals = Y_residuals_permuted[finite_Y_permuted]
        covariance = (reduced_X_residuals * reduced_Y_residuals).sum()
        # The denominator will be different on each permutation
        denominator = np.sqrt(
            sum(reduced_X_residuals**2) * sum(reduced_Y_residuals**2)
        )
        correlations[i] = covariance / denominator
    return correlations


def stochastic_test(m, n, X_residuals, Y_residuals_as_matrix):
    Y_residuals_permuted = np.zeros((m**2 - m) // 2)
    covariances = np.zeros(n)
    order = np.arange(m)
    for i in range(1, n):
        np.random.shuffle(order)
        Y_residuals_as_matrix_permuted = Y_residuals_as_matrix[order, :][:, order]
        spatial.distance._distance_wrap.to_vector_from_squareform_wrap(
            Y_residuals_as_matrix_permuted, Y_residuals_permuted
        )
        covariances[i] = (X_residuals * Y_residuals_permuted).sum()
    denominator = np.sqrt(sum(X_residuals**2) * sum(Y_residuals_permuted**2))
    return covariances / denominator


def stochastic_test_with_nans(m, n, X, Y_residuals_as_matrix):
    Y_residuals_permuted = np.zeros((m**2 - m) // 2)
    correlations = np.zeros(n)
    order = np.arange(m)
    for i in range(1, n):
        np.random.shuffle(order)
        Y_residuals_as_matrix_permuted = Y_residuals_as_matrix[order, :][:, order]
        spatial.distance._distance_wrap.to_vector_from_squareform_wrap(
            Y_residuals_as_matrix_permuted, Y_residuals_permuted
        )
        # Since on each permutation we will be ignoring different values in X,
        # the X_residuals need to be recomputed each time depending on which
        # values in permuted Y are finite.
        finite_Y_permuted = np.isfinite(Y_residuals_permuted)
        reduced_X = X[finite_Y_permuted]
        reduced_X_residuals = reduced_X - reduced_X.mean()
        reduced_Y_residuals = Y_residuals_permuted[finite_Y_permuted]
        covariance = (reduced_X_residuals * reduced_Y_residuals).sum()
        # The denominator will be different on each permutation
        denominator = np.sqrt(
            sum(reduced_X_residuals**2) * sum(reduced_Y_residuals**2)
        )
        correlations[i] = covariance / denominator
    return correlations



import numpy as np
import mantel


# Example with condendensed distance matrices

dists1 = [0.2, 0.4, 0.3, 0.6, 0.9, 0.4]
dists2 = [0.3, 0.3, 0.2, 0.7, 0.8, 0.3]

result = mantel.test(dists1, dists2, method="pearson", tail="upper")
print(result)


# Example with redundant distance matrices

dists1 = [[0.0,0.2,0.4,0.3],
          [0.2,0.0,0.6,0.9],
          [0.4,0.6,0.0,0.4],
          [0.3,0.9,0.4,0.0]]
dists2 = [[0.0,0.3,0.3,0.2],
          [0.3,0.0,0.7,0.8],
          [0.3,0.7,0.0,0.3],
          [0.2,0.8,0.3,0.0]]

result = mantel.test(dists1, dists2, method="pearson", tail="upper")
print(result)


# Example with random data (implying no correlation)

n_objects = 27
n_distances = (n_objects**2 - n_objects) // 2

dists1 = np.random.random(n_distances)
dists2 = np.random.random(n_distances)

result = mantel.test(dists1, dists2)

print(result.r)
print(result.p)
print(result.z)

print(result.p < 0.05)

print(result.correlations)
print(result.mean)
print(result.std)


# Plotting example (requires matplotlib)

fig, axis = mantel.plot(result)
fig.savefig('example.svg')



import setuptools

with open("README.md", encoding="utf-8") as file:
    long_description = file.read()

setuptools.setup(
    name="mantel",
    use_scm_version={
        "write_to": "mantel/_version.py",
        "write_to_template": '__version__ = "{version}"',
        "fallback_version": "???",
    },
    author="Jon Carr",
    author_email="jcarr@sissa.it",
    description="Python implementation of the Mantel test, a significance test of the correlation between two distance matrices",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jwcarr/mantel",
    license="MIT",
    packages=["mantel"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Mathematics",
    ],
    python_requires=">=3.6",
    install_requires=["numpy>=1.10", "scipy>=1.0"],
    setup_requires=["setuptools_scm"],
)
